# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# cSpell: ignore Aroba elems OPEXPR Predef prio Verif
# pylint: disable=invalid-name, too-many-lines, no-else-return, too-many-branches,
# pylint: disable=inconsistent-return-statements, too-many-return-statements
# pylint: disable=singleton-comparison, too-many-locals, too-many-statements

"""
The PyOfAst module transforms F# AST into Python ansys.scadeone.core.swan classes.
"""

from typing import cast

# pylint: disable-next=import-error
from ANSYS.SONE.Infrastructure.Services.Serialization.BNF.Parsing import Ast, Raw  # type:ignore

from ansys.scadeone.core.common.exception import ScadeOneException
from ansys.scadeone.core.common.storage import SwanString
import ansys.scadeone.core.swan as S
from ansys.scadeone.core.swan.pragmas import PragmaParser

from .parser import Parser


def getValueOf(option):
    """Help to get value from 't option"""
    return option.Value if option else None


def getMarkup(raw):
    return Raw.getMarkup(raw)


def getProtectedString(raw):
    return Raw.getIndentedRawString(raw)


def protectedItemOfAst(raw):
    return S.ProtectedItem(getProtectedString(raw), getMarkup(raw))


def getPragmas(ast_pragmas):
    return [S.Pragma(p) for p in ast_pragmas]


# Identifiers
# ============================================================
def identifierOfAst(ast):
    id = Ast.idName(ast)
    pragmas = getPragmas(Ast.idPragmas(ast))
    return S.Identifier(id, pragmas)


def pathIdentifierOfAst(pathId):
    ids = [identifierOfAst(id) for id in pathId]
    return S.PathIdentifier(ids)


def pathIdentifierOrRawOfAst(pathId):
    if pathId.IsPIOfId:
        return pathIdentifierOfAst(pathId.Item1)
    return S.PathIdentifier(getProtectedString(pathId.Item))


def stringOfStringWithSP(ast):
    return ast.StringData


def instanceIdOfAst(ast) -> str:
    if ast.IsInstanceIdSelf:
        return "self"
    return stringOfStringWithSP(ast.Item)


def nameOfAst(ast) -> str:
    # skip '
    return stringOfStringWithSP(ast)[1:]


def luidOfAst(ast):
    return S.Luid(stringOfStringWithSP(ast))


def lunumOfAst(ast):
    return S.Lunum(stringOfStringWithSP(ast))


# Expressions
# ============================================================


# arithmetic & logical operators
# ------------------------------
def unaryOfOfAst(ast):
    if ast.IsUMinus:
        return S.UnaryOp.Minus
    elif ast.IsUPlus:
        return S.UnaryOp.Plus
    elif ast.IsULnot:
        return S.UnaryOp.Lnot
    elif ast.IsUNot:
        return S.UnaryOp.Not
    elif ast.IsUPre:
        return S.UnaryOp.Pre


def binaryOpOfAst(ast):
    if ast.IsBPlus:
        return S.BinaryOp.Plus
    elif ast.IsBMinus:
        return S.BinaryOp.Minus
    elif ast.IsBMult:
        return S.BinaryOp.Mult
    elif ast.IsBSlash:
        return S.BinaryOp.Slash
    elif ast.IsBMod:
        return S.BinaryOp.Mod
    elif ast.IsBLand:
        return S.BinaryOp.Land
    elif ast.IsBLor:
        return S.BinaryOp.Lor
    elif ast.IsBLxor:
        return S.BinaryOp.Lxor
    elif ast.IsBLsl:
        return S.BinaryOp.Lsl
    elif ast.IsBLsr:
        return S.BinaryOp.Lsr
    elif ast.IsBEqual:
        return S.BinaryOp.Equal
    elif ast.IsBDiff:
        return S.BinaryOp.Diff
    elif ast.IsBLt:
        return S.BinaryOp.Lt
    elif ast.IsBGt:
        return S.BinaryOp.Gt
    elif ast.IsBLeq:
        return S.BinaryOp.Leq
    elif ast.IsBGeq:
        return S.BinaryOp.Geq
    elif ast.IsBAnd:
        return S.BinaryOp.And
    elif ast.IsBOr:
        return S.BinaryOp.Or
    elif ast.IsBXor:
        return S.BinaryOp.Xor
    elif ast.IsBArrow:
        return S.BinaryOp.Arrow
    elif ast.IsBPre:
        return S.BinaryOp.Pre
    elif ast.IsBAroba:
        return S.BinaryOp.Concat


# label, group item
# -----------------
def labelOrIndexOfAst(ast):
    if ast.IsIndex:
        expr = exprOrRawOfAst(ast.Item)
        return S.LabelOrIndex(expr)
    id = identifierOfAst(ast.Item)
    return S.LabelOrIndex(id)


def groupItemOfAst(ast):
    expr = exprOrRawOfAst(ast.Item) if ast.IsGroupItemExpr else exprOrRawOfAst(ast.Item2)
    label = identifierOfAst(ast.Item1) if ast.IsGroupItemLabelExpr else None
    return S.GroupItem(expr, label)


def groupOfAst(ast):
    items = [groupItemOfAst(item) for item in ast]
    return S.Group(items)


# modifiers, patterns
# -------------------
def modifierOfAst(ast):
    new_value = exprOrRawOfAst(ast.Item2)
    if ast.IsModifierRaw:
        return S.Modifier(getProtectedString(ast.Item1), new_value)
    indices = [labelOrIndexOfAst(item) for item in ast.Item1]
    return S.Modifier(indices, new_value)


def casePatternsOfAst(cases):
    def caseOfAst(pattern, expr):
        p_obj = patternOrRawOfAst(pattern)
        e_obj = exprOrRawOfAst(expr)
        return S.CaseBranch(p_obj, e_obj)

    patterns = [caseOfAst(c.Item1, c.Item2) for c in cases]
    return patterns


def patternOrRawOfAst(pattern):
    if pattern.IsPRaw:
        return S.ProtectedPattern(getProtectedString(pattern.Item))

    pattern = pattern.Item1
    if pattern.IsPId:
        tag = pathIdentifierOfAst(pattern.Item)
        return S.PathIdPattern(tag)

    if pattern.IsPVariant:
        tag = pathIdentifierOfAst(pattern.Item)
        return S.VariantPattern(tag, underscore=True)

    if pattern.IsPVariantCapture:
        tag = pathIdentifierOfAst(pattern.Item1)
        if id := getValueOf(pattern.Item2):
            return S.VariantPattern(tag, identifierOfAst(id))
        return S.VariantPattern(tag)

    if pattern.IsPChar:
        return S.CharPattern(pattern.Item)

    if pattern.IsPInt:
        return S.IntPattern(pattern.Item2, pattern.Item1)

    if pattern.IsPBool:
        return S.BoolPattern(pattern.Item)

    if pattern.IsPUscore:
        return S.UnderscorePattern()

    if pattern.IsPDefault:
        return S.DefaultPattern()


# Renamings
# ---------
def renamingOfAst(ast):
    if ast.IsRenamingByPos:  # of string * bool * Id option
        index = S.Literal(ast.Item1)
        is_shortcut = ast.Item2
        if renaming := getValueOf(ast.Item3):
            renaming = identifierOfAst(renaming)
        return S.GroupRenaming(index, renaming, is_shortcut)

    if ast.IsRenamingByName:  #  of Id * bool * Id option
        index = identifierOfAst(ast.Item1)
        is_shortcut = ast.Item2
        if renaming := getValueOf(ast.Item3):
            renaming = identifierOfAst(renaming)
        return S.GroupRenaming(index, renaming, is_shortcut)

    if ast.IsRenamingRaw:  # of SourcePosition.t
        markup = getMarkup(ast.Item)
        content = getProtectedString(ast.Item)
        return S.ProtectedGroupRenaming(content, markup)


def groupAdaptationOfAst(ast):
    renamings = [renamingOfAst(ren) for ren in ast.GRenaming]
    return S.GroupAdaptation(renamings)


# Clock expression
# ----------------
def clockExprOfAst(ast):
    if ast.IsClockId:  # of Id
        return S.ClockExpr(identifierOfAst(ast.Item))
    if ast.IsClockNotId:  # of Id
        return S.ClockExpr(identifierOfAst(ast.Item), is_not=True)
    if ast.IsClockMatch:  # of Id * PatternOrRaw
        pattern = patternOrRawOfAst(ast.Item2)
        return S.ClockExpr(identifierOfAst(ast.Item1), pattern=pattern)


# Forward expression
# ~~~~~~~~~~~~~~~~~~~
def forwardLHSofAst(ast):
    if ast.IsFId:  # of Id
        return S.ForwardLHS(identifierOfAst(ast.Item))
    # FLhsArray of ForwardLhs
    return S.ForwardLHS(forwardLHSofAst(ast.Item))


def forwardElement(ast):
    lhs = forwardLHSofAst(ast.Item1)
    expr = expressionOfAst(ast.Item2)
    return S.ForwardElement(lhs, expr)


def forwardDimOfAst(ast):
    if ast.IsFDim:  # of Expr * SourcePosition.t
        return S.ForwardDim(expressionOfAst(ast.Item1))

    if ast.IsFDimWith:  # of Expr * Id option * (ForwardLhs * Expr) list * SourcePosition.t
        expr = expressionOfAst(ast.Item1)
        if id := getValueOf(ast.Item2):
            id = identifierOfAst(id)
        elems = [forwardElement(elem) for elem in ast.Item3]
        return S.ForwardDim(expr, id, elems)

    # FRaw of Raw.t
    data = getProtectedString(ast.Item)
    return S.ForwardDim(protected=data)


def forwardBodyOfAst(ast):
    sections = [scopeSectionOfAst(sec) for sec in ast.FScopeSections]
    if until := getValueOf(ast.FUntilCondition):
        until = exprOrRawOfAst(until)
    if unless := getValueOf(ast.FUnlessCondition):
        unless = exprOrRawOfAst(unless)

    return S.ForwardBody(sections, unless, until)


def forwardLastDefaultOfAst(ast):
    if ast.IsFLast:  # of Expr
        return S.ForwardLastDefault(last=expressionOfAst(ast.Item))

    if ast.IsFDefault:  # of Expr
        return S.ForwardLastDefault(default=expressionOfAst(ast.Item))

    if ast.IsFLastPlusDefault:  # of Expr * Expr
        return S.ForwardLastDefault(
            last=expressionOfAst(ast.Item1), default=expressionOfAst(ast.Item2)
        )
    # ast.IsFLastAndDefault: # of Expr
    return S.ForwardLastDefault(shared=expressionOfAst(ast.Item))


def forwardItemClauseOfAst(ast):
    id = identifierOfAst(ast.Item1)
    if last_default := getValueOf(ast.Item2):
        last_default = forwardLastDefaultOfAst(last_default)
    return S.ForwardItemClause(id, last_default)


def forwardArrayClauseOfAst(ast):
    if ast.IsFItemClause:  # of ForwardItemClause
        clause = forwardItemClauseOfAst(ast.Item)
    else:  # ast.IsFArrayClause  of ForwardArrayClause
        clause = forwardArrayClauseOfAst(ast.Item)
    return S.ForwardArrayClause(clause)


def forwardReturnOfAst(ast):
    if ast.IsFRetItemClause:  # of ForwardItemClause * SourcePosition.t
        clause = forwardItemClauseOfAst(ast.Item1)
        return S.ForwardReturnItemClause(clause)
    if ast.IsFRetArrayClause:  # of Id option * ForwardArrayClause * SourcePosition.t
        if id := getValueOf(ast.Item1):
            id = identifierOfAst(id)
        clause = forwardArrayClauseOfAst(ast.Item2)
        return S.ForwardReturnArrayClause(clause, id)
    if ast.IsFRetRaw:  # of Raw.t
        return S.ProtectedForwardReturnItem(getProtectedString(ast.Item))


def forwardOfAst(ast):
    # Luid option * ForwardState * ForwardDim list * ForwardBody * ForwardReturnsItem list
    if luid := getValueOf(ast.Item1):
        luid = luidOfAst(luid)

    if ast.Item2.IsFNone:
        state = S.ForwardState.Nothing
    elif ast.Item2.IsFRestart:
        state = S.ForwardState.Restart
    else:  # IsFResume
        state = S.ForwardState.Resume

    dims = [forwardDimOfAst(dim) for dim in ast.Item3]
    body = forwardBodyOfAst(ast.Item4)
    returns = [forwardReturnOfAst(ret) for ret in ast.Item5]

    return S.Forward(state, dims, body, returns, luid)


# Operator instance & expressions
# -------------------------------
def iteratorOfAst(ast, operator):
    # pylint: disable=possibly-used-before-assignment

    if ast.IsIMap:
        kind = S.IteratorKind.Map
    elif ast.IsIFold:
        kind = S.IteratorKind.Fold
    elif ast.IsIMapi:
        kind = S.IteratorKind.Mapi
    elif ast.IsIFoldi:
        kind = S.IteratorKind.Foldi
    elif ast.IsIMapfold:  # of int
        kind = S.IteratorKind.Mapfold
    elif ast.IsIMapfoldi:  # of int
        kind = S.IteratorKind.Mapfoldi

    return S.Iterator(kind, operator)


def optGroupItemOfAst(ast):
    group_item = groupItemOfAst(ast.Item) if ast.IsOGroupItem else None
    return S.OptGroupItem(group_item)


def operatorExprWithSPOfAst(ast):
    return operatorExprOfAst(ast.OEOpExpr)


def operatorExprOfAst(ast):
    if ast.IsOIterator:  # Iterator * Operator
        operator = operatorBaseOfAst(ast.Item2)
        return iteratorOfAst(ast.Item1, operator)

    if ast.IsOActivateClock:  # Operator * ClockExpr
        operator = operatorBaseOfAst(ast.Item1)
        clock = clockExprOfAst(ast.Item2)
        return S.ActivateClock(operator, clock)

    if ast.IsOActivateCondition:  # Operator * ExprOrRaw * bool * ExprOrRaw
        operator = operatorBaseOfAst(ast.Item1)
        cond = exprOrRawOfAst(ast.Item2)
        is_last = ast.Item3
        default = exprOrRawOfAst(ast.Item4)
        return S.ActivateEvery(operator, cond, is_last, default)

    if ast.IsORestart:  # Operator * ExprOrRaw
        operator = operatorBaseOfAst(ast.Item1)
        condition = exprOrRawOfAst(ast.Item2)
        return S.Restart(operator, condition)

    if ast.IsOLambdaDataDef:  # bool * VarOrRaw list * VarOrRaw list * ScopeDefinition
        is_node = ast.Item1
        inputs = [varDeclOfAst(sig) for sig in ast.Item2]
        outputs = [varDeclOfAst(sig) for sig in ast.Item3]
        data_def = scopeOfAst(ast.Item4)
        return S.AnonymousOperatorWithDataDefinition(is_node, inputs, outputs, data_def)

    if ast.IsOLambdaScopes:  # bool * Id list * ScopeSection list * ExprOrRaw
        is_node = ast.Item1
        params = [identifierOfAst(id) for id in ast.Item2]
        sections = [scopeSectionOfAst(scope) for scope in ast.Item3]
        expr = exprOrRawOfAst(ast.Item4)
        return S.AnonymousOperatorWithExpression(is_node, params, sections, expr)

    if ast.IsOPartial:  # Operator * OptGroupItem list
        operator = operatorBaseOfAst(ast.Item1)
        partial_group = [optGroupItemOfAst(item) for item in ast.Item2]
        return S.Partial(operator, partial_group)

    if ast.IsONary:  # BinaryOp // ONary is a subset of BinaryOp
        nary = ast.Item
        if nary.IsBPlus:
            return S.NAryOperator(S.NaryOp.Plus)
        if nary.IsBMult:
            return S.NAryOperator(S.NaryOp.Mult)
        if nary.IsBLand:
            return S.NAryOperator(S.NaryOp.Land)
        if nary.IsBLor:
            return S.NAryOperator(S.NaryOp.Lor)
        if nary.IsBAnd:
            return S.NAryOperator(S.NaryOp.And)
        if nary.IsBOr:
            return S.NAryOperator(S.NaryOp.Or)
        if nary.IsBXor:
            return S.NAryOperator(S.NaryOp.Xor)
        if nary.IsBAroba:
            return S.NAryOperator(S.NaryOp.Concat)
    if ast.IsOSource:
        return S.Source(ast.Item)
    if ast.IsOOracle:
        return S.Oracle(ast.Item)


def operatorPrefixOfAst(ast, sizes, pragmas):
    if ast.IsOPathId:  # PathId
        id = pathIdentifierOfAst(ast.Item)
        return S.PathIdOpCall(id, sizes, pragmas)

    if ast.IsOPrefixPrimitive:  # PrefixPrimitive
        prefix = ast.Item
        if prefix.IsFlatten:
            kind = S.PrefixPrimitiveKind.Flatten
        elif prefix.IsPack:
            kind = S.PrefixPrimitiveKind.Pack
        elif prefix.IsReverse:
            kind = S.PrefixPrimitiveKind.Reverse
        else:  # ast.IsTranspose
            kind = S.PrefixPrimitiveKind.Transpose

        if kind != S.PrefixPrimitiveKind.Transpose:
            return S.PrefixPrimitive(kind, sizes)
        # Transpose
        index = prefix.Item
        if index.IsTSList:
            params = [index for index in index.Item1]
        else:
            params = getProtectedString(index.Item)
        return S.Transpose(params, sizes)

    if ast.IsORawPrefix or ast.IsORawOpExpr:
        # Protected content, find what it is.
        markup = getMarkup(ast.Item)
        source = getProtectedString(ast.Item)
        if markup == "text":
            # text markup is used for an operator expression
            # which is not between parentheses and not an operator instance.
            origin = Parser.get_source().name
            swan = SwanString(f"{source}", origin)
            op_block = Parser.get_current_parser().operator_block(swan)
            op_block.is_text = True
            # sizes may be defined outside of the operator block (form)
            # Note: having both sizes in text and in the operator block is not allowed
            # and SLS raises an error => we don't check it here, as it is a syntax error
            if sizes:
                op_block._sizes = sizes
            return op_block

        if markup == "op_expr":
            # ORawPrefix is returned for: LP RAW_OPEXPR RP
            # which is an operator expression between parentheses.
            origin = Parser.get_source().name
            swan = SwanString(f"{source}", origin)
            op_expr = Parser.get_current_parser().op_expr(swan)
            op_expr.is_op_expr = True
            return S.PrefixOperatorExpression(op_expr, sizes)

        return S.ProtectedOpExpr(source, markup)

    # op_expr OOperatorExpr
    op_expr = operatorExprWithSPOfAst(ast.Item)
    return S.PrefixOperatorExpression(op_expr, sizes)


def operatorBaseOfAst(ast):
    sizes = [exprOrRawOfAst(sz) for sz in ast.CallSize]
    pragmas = getPragmas(ast.CallPragmas)
    return operatorPrefixOfAst(ast.CallOp, sizes, pragmas)


# Expressions or raw
# ------------------
def exprOrRawOfAst(ast):
    if ast.IsExprWithSP:  # of Expr * SourcePosition.t
        return expressionOfAst(ast.Item1)
    return S.ProtectedExpr(getProtectedString(ast.Item), getMarkup(ast.Item))


def expressionOfAst(ast):
    if ast.IsEId:  #  of PathId
        path_id = pathIdentifierOfAst(ast.Item)
        return S.PathIdExpr(path_id)

    elif ast.IsELast:  #  of Name
        return S.LastExpr(S.Identifier(nameOfAst(ast.Item), is_name=True))

    elif ast.IsEBoolLiteral:  #  of bool
        return S.Literal("true" if ast.Item else "false")

    elif ast.IsECharLiteral:  #  of string
        return S.Literal(ast.Item)

    elif ast.IsENumLiteral:  #  of string
        return S.Literal(ast.Item)

    elif ast.IsEUnaryOp:  #  of UnaryOp * ExprOrRaw
        return S.UnaryExpr(unaryOfOfAst(ast.Item1), exprOrRawOfAst(ast.Item2))

    elif ast.IsEBinaryOp:  #  of BinaryOp * ExprOrRaw * ExprOrRaw
        return S.BinaryExpr(
            binaryOpOfAst(ast.Item1),
            exprOrRawOfAst(ast.Item2),
            exprOrRawOfAst(ast.Item3),
        )

    elif ast.IsEWhenClock:  #  of ExprOrRaw * ClockExpr
        expr = exprOrRawOfAst(ast.Item1)
        ck = clockExprOfAst(ast.Item2)
        return S.WhenClockExpr(expr, ck)

    elif ast.IsEWhenMatch:  #  of ExprOrRaw * PathId
        expr = exprOrRawOfAst(ast.Item1)
        match = pathIdentifierOfAst(ast.Item2)
        return S.WhenMatchExpr(expr, match)

    elif ast.IsECast:  #  of ExprOrRaw * TypeExprOrRaw
        expr = exprOrRawOfAst(ast.Item1)
        type = typeOrRawOfAst(ast.Item2)
        return S.NumericCast(expr, type)

    elif ast.IsEGroup:  #  of Group
        items = [groupItemOfAst(item) for item in ast.Item]
        return S.GroupConstructor(S.Group(items))

    elif ast.IsEGroupAdapt:  #  of ExprOrRaw * GroupAdaptation
        expr = exprOrRawOfAst(ast.Item1)
        adaptation = groupAdaptationOfAst(ast.Item2)
        return S.GroupProjection(expr, adaptation)

    ## Composite
    elif ast.IsEStaticProj:  #  of ExprOrRaw * LabelOrIndex
        expr = exprOrRawOfAst(ast.Item1)
        labelOrIndex = labelOrIndexOfAst(ast.Item2)
        if labelOrIndex.is_label:
            return S.StructProjection(expr, labelOrIndex)
        else:
            return S.ArrayProjection(expr, labelOrIndex)

    elif ast.IsEMkGroup:  #  of PathIdOrRaw * ExprOrRaw
        name = pathIdentifierOrRawOfAst(ast.Item1)
        expr = exprOrRawOfAst(ast.Item2)
        return S.StructDestructor(name, expr)

    elif ast.IsESlice:  #  of ExprOrRaw * ExprOrRaw * ExprOrRaw
        expr = exprOrRawOfAst(ast.Item1)
        start = exprOrRawOfAst(ast.Item2)
        end = exprOrRawOfAst(ast.Item3)
        return S.Slice(expr, start, end)

    elif ast.IsEDynProj:  #  of ExprOrRaw * LabelOrIndex list * ExprOrRaw (* default *)
        expr = exprOrRawOfAst(ast.Item1)
        indices = [labelOrIndexOfAst(item) for item in ast.Item2]
        default = exprOrRawOfAst(ast.Item3)
        return S.ProjectionWithDefault(expr, indices, default)

    elif ast.IsEMkArray:  #  of ExprOrRaw * ExprOrRaw
        expr = exprOrRawOfAst(ast.Item1)
        size = exprOrRawOfAst(ast.Item2)
        return S.ArrayRepetition(expr, size)

    elif ast.IsEMkArrayGroup:  #  of Group
        return S.ArrayConstructor(groupOfAst(ast.Item))

    elif ast.IsEMkStruct:  #  of Group * PathIdOrRaw option
        group = groupOfAst(ast.Item1)
        if id := getValueOf(ast.Item2):
            id = pathIdentifierOrRawOfAst(id)
        return S.StructConstructor(group, id)

    elif ast.IsEVariant:  #  of PathIdOrRaw * Group
        tag = pathIdentifierOrRawOfAst(ast.Item1)
        group = groupOfAst(ast.Item2)
        return S.VariantValue(tag, group)

    elif ast.IsEMkCopy:  #  of ExprOrRaw * Modifier list
        expr = exprOrRawOfAst(ast.Item1)
        modifiers = [modifierOfAst(item) for item in ast.Item2]
        return S.FunctionalUpdate(expr, modifiers)

    ## Switch
    elif ast.IsEIfte:  #  of ExprOrRaw * ExprOrRaw * ExprOrRaw
        cond_expr = exprOrRawOfAst(ast.Item1)
        then_expr = exprOrRawOfAst(ast.Item2)
        else_expr = exprOrRawOfAst(ast.Item3)
        return S.IfteExpr(cond_expr, then_expr, else_expr)

    elif ast.IsECase:  #  of ExprOrRaw * (PatternOrRaw * ExprOrRaw) list
        expr = exprOrRawOfAst(ast.Item1)
        patterns = casePatternsOfAst(ast.Item2)
        return S.CaseExpr(expr, patterns)

    ## OpCalls & Ports
    elif ast.IsEOpCall:  #  of OperatorInstance * Group
        params = groupOfAst(ast.Item2)
        if luid := getValueOf(ast.Item1.OIInstance):
            luid = luidOfAst(luid)
        operator = operatorBaseOfAst(ast.Item1.OIOperator)
        return S.OperatorInstance(operator, params, luid)

    elif ast.IsEPort:  #  of Port
        return portOfAst(ast.Item)

    ## Forward loops
    elif ast.IsEForward:
        #  of Luid option * ForwardState * ForwardDim list
        # * ForwardBody * ForwardReturnsItem list
        return forwardOfAst(ast)

    elif ast.IsEWindow:  #  of ExprOrRaw * Group * Group
        expr = exprOrRawOfAst(ast.Item1)
        params = groupOfAst(ast.Item2)
        init = groupOfAst(ast.Item3)
        return S.Window(expr, params, init)

    elif ast.IsEMerge:  #  of Group list
        params = [groupOfAst(group) for group in ast.Item]
        return S.Merge(params)


def portOfAst(ast):
    if ast.IsInstanceIdLunum:
        lunum = lunumOfAst(ast.Item)
        return S.PortExpr(lunum=lunum)
    if ast.IsInstanceIdLuid:
        luid = luidOfAst(ast.Item)
        return S.PortExpr(luid=luid)
    elif ast.IsInstanceIdSelf:
        return S.PortExpr(is_self=True)
    else:
        raise ScadeOneException("internal error, unexpected instance id")


# Type Expressions
# ============================================================
def predefinedTypeOfAst(ast):
    if ast.IsBool:
        return S.BoolType()
    elif ast.IsChar:
        return S.CharType()
    elif ast.IsInt8:
        return S.Int8Type()
    elif ast.IsInt16:
        return S.Int16Type()
    elif ast.IsInt32:
        return S.Int32Type()
    elif ast.IsInt64:
        return S.Int64Type()
    elif ast.IsUint8:
        return S.Uint8Type()
    elif ast.IsUint16:
        return S.Uint16Type()
    elif ast.IsUint32:
        return S.Uint32Type()
    elif ast.IsUint64:
        return S.Uint64Type()
    elif ast.IsFloat32:
        return S.Float32Type()
    elif ast.IsFloat64:
        return S.Float64Type()


def typeExpressionOfAst(ast):
    if ast.IsTPredefinedType:  # of PredefType
        return predefinedTypeOfAst(ast.Item)

    elif ast.IsTSizedSigned:  # of Expr
        expr = expressionOfAst(ast.Item)
        return S.SizedTypeExpression(expr, True)

    elif ast.IsTSizedUnsigned:  # of Expr
        expr = expressionOfAst(ast.Item)
        return S.SizedTypeExpression(expr, False)

    elif ast.IsTAlias:  # of PathId
        path_id = pathIdentifierOfAst(ast.Item)
        return S.TypeReferenceExpression(path_id)

    elif ast.IsTVar:  # of StringWithSourcePosition
        var = S.Identifier(nameOfAst(ast.Item), is_name=True)
        return S.VariableTypeExpression(var)

    elif ast.IsTArray:  # of TypeExpr * Expr
        type = typeExpressionOfAst(ast.Item1)
        size = expressionOfAst(ast.Item2)
        return S.ArrayTypeExpression(type, size)


def typeOrRawOfAst(ast):
    if ast.IsRawTypeExpr:
        return S.ProtectedTypeExpr(getProtectedString(ast.Item))
    return typeExpressionOfAst(ast.Item1)


# Declarations
# ============================================================


# Global declarations
# ------------------------------------------------------------
def constDecl(ast):
    id = identifierOfAst(ast.ConstId)
    if value := getValueOf(ast.ConstDefinition):
        value = expressionOfAst(value)
    if type := getValueOf(ast.ConstType):
        type = typeExpressionOfAst(type)
    return S.ConstDecl(id, type, value)


def sensorDecl(ast):
    id = identifierOfAst(ast.SensorId)
    type = typeExpressionOfAst(ast.SensorType)
    return S.SensorDecl(id, type)


def structFieldsOfAst(ast):
    def field(ast):
        id = identifierOfAst(ast.Item1)
        type = typeExpressionOfAst(ast.Item2)
        return S.StructField(id, type)

    fields = [field(f) for f in ast]
    return fields


def typeDecl(ast):
    id = identifierOfAst(ast.TypeId)
    if ast.TypeDef.IsTDefNone:
        return S.TypeDecl(id)

    elif ast.TypeDef.IsTDefExpr:  # of TypeExpr
        type_expr = typeExpressionOfAst(ast.TypeDef.Item)
        type_def = S.ExprTypeDefinition(type_expr)
        return S.TypeDecl(id, type_def)

    elif ast.TypeDef.IsTDefEnum:  # of Id list
        tags = [identifierOfAst(t) for t in ast.TypeDef.Item]
        enum_decl = S.EnumTypeDefinition(tags)
        return S.TypeDecl(id, enum_decl)

    elif ast.TypeDef.IsTDefVariant:  # of TypeVariant list

        def variantOfAst(ast):
            tag = identifierOfAst(ast.Item1)
            if ast.Item2.IsVTSimple:
                return S.VariantSimple(tag)
            elif ast.Item2.IsVTTypeExpr:
                type_expr = typeExpressionOfAst(ast.Item2.Item)
                return S.VariantTypeExpr(tag, type_expr)
            fields = structFieldsOfAst(ast.Item2.Item)
            return S.VariantStruct(tag, fields)

        tags = [variantOfAst(v) for v in ast.TypeDef.Item]
        variant_decl = S.VariantTypeDefinition(tags)
        return S.TypeDecl(id, variant_decl)

    elif ast.TypeDef.IsTDefStruct:  # of StructField list
        # StructField = Id * TypeExpr
        fields = structFieldsOfAst(ast.TypeDef.Item)
        struct_decl = S.StructTypeDefinition(fields)
        return S.TypeDecl(id, struct_decl)


def useDecl(ast) -> S.UseDirective:
    path_id = pathIdentifierOfAst(ast.UPath)
    if alias := getValueOf(ast.UAs):
        alias = identifierOfAst(alias)
    return S.UseDirective(path_id, alias)


def groupDecl(ast):
    id = identifierOfAst(ast.GroupId)
    type = groupTypeExprOfAst(ast.GroupType)
    return S.GroupDecl(id, type)


def groupTypeExprOfAst(ast) -> S.GroupTypeExpression:
    if ast.IsGTypeExpr:
        type = typeExpressionOfAst(ast.Item)
        return S.TypeGroupTypeExpression(type)
    positional = [groupTypeExprOfAst(pos) for pos in ast.Item1]

    def namedGroupExprOfAst(ast):
        id = identifierOfAst(ast.Item1)
        type = groupTypeExprOfAst(ast.Item2)
        return S.NamedGroupTypeExpression(id, type)

    named = [namedGroupExprOfAst(named) for named in ast.Item2]
    return S.GroupTypeExpressionList(positional, named)


# Operator & Signature declarations
# ------------------------------------------------------------
def numericKindOfAst(ast):
    if ast.IsNumeric:
        return S.NumericKind.Numeric
    if ast.IsInteger:
        return S.NumericKind.Integer
    if ast.IsSigned:
        return S.NumericKind.Signed
    if ast.IsUnsigned:
        return S.NumericKind.Unsigned
    if ast.IsFloat:
        return S.NumericKind.Float


def constraintOfAst(ast) -> S.TypeConstraint:
    num_kind = numericKindOfAst(ast.Item2)
    if ast.IsTCRaw:
        return S.TypeConstraint(getProtectedString(ast.Item1), num_kind)
    type_vars = [typeExpressionOfAst(tv) for tv in ast.Item1]
    return S.TypeConstraint(type_vars, num_kind)


def isProbePragma(p: S.Pragma):
    d = PragmaParser.extract(p.pragma)
    return d == ("cg", "probe")


def varDeclOfAst(ast) -> S.Variable:
    if ast.IsRawVar:
        return S.ProtectedVariable(getProtectedString(ast.Item))
    var_decl = ast.Item1
    id = identifierOfAst(var_decl.VarId)
    is_clock = var_decl.VarIsClock
    is_probe = var_decl.VarIsProbe
    if is_probe:
        id._pragmas = [p for p in id.pragmas if not isProbePragma(p)]
    if type := getValueOf(var_decl.VarType):
        type = groupTypeExprOfAst(type)
    if when := getValueOf(var_decl.VarWhen):
        when = clockExprOfAst(when)
    if default := getValueOf(var_decl.VarDefault):
        default = expressionOfAst(default)
    if last := getValueOf(var_decl.VarLast):
        last = expressionOfAst(last)

    return S.VarDecl(id, is_clock, is_probe, type, when, default, last)


def signatureElementsOfAst(ast):
    inline = ast.OpInline
    kind = ast.OpNode
    name = S.Identifier(stringOfStringWithSP(ast.OpId))
    inputs = [varDeclOfAst(sig) for sig in ast.OpInputs]
    for sig in inputs:
        sig.is_input = True
    outputs = [varDeclOfAst(sig) for sig in ast.OpOutputs]
    for sig in outputs:
        sig.is_output = True
    sizes = [identifierOfAst(id) for id in ast.OpSizes]
    constraints = [constraintOfAst(ct) for ct in ast.OpConstraints]
    if specialization := getValueOf(ast.OpSpecialization):
        specialization = pathIdentifierOrRawOfAst(specialization)
    pragmas = getPragmas(ast.OpPragmas)
    return (
        inline,
        kind,
        name,
        inputs,
        outputs,
        sizes,
        constraints,
        specialization,
        pragmas,
    )


def signatureOfAst(ast):
    (
        inline,
        kind,
        name,
        inputs,
        outputs,
        sizes,
        constraints,
        specialization,
        pragmas,
    ) = signatureElementsOfAst(ast)

    return S.Signature(
        id=name,
        is_inlined=inline,
        is_node=kind,
        inputs=inputs,
        outputs=outputs,
        sizes=sizes,
        constraints=constraints,
        specialization=specialization,
        pragmas=pragmas,
    )


def emissionBodyOfAst(ast):
    flows = [S.Identifier(nameOfAst(sig), is_name=True) for sig in ast.ESignals]
    if condition := getValueOf(ast.EExpr):
        condition = expressionOfAst(condition)
    if luid := getValueOf(ast.ELuid):
        luid = luidOfAst(luid)
    return S.EmissionBody(flows, condition, luid)


# Equations
# ------------------------------------------------------------


def lhsOfAst(ast):
    if ast.IsLhsId:
        return S.LHSItem(identifierOfAst(ast.Item))
    return S.LHSItem()


def equationLhsOfAst(ast):
    if ast.IsLhsEmpty:
        return S.EquationLHS([])
    lhs_items = [lhsOfAst(lhs) for lhs in ast.Item]
    return S.EquationLHS(lhs_items, ast.IsLhsWithRest)


def equationOfAst(ast):
    if ast.IsEquation:  # of Luid option * Lhs * Expr * SourcePosition.t
        if luid := getValueOf(ast.Item1):
            luid = luidOfAst(luid)
        lhs = equationLhsOfAst(ast.Item2)
        expr = expressionOfAst(ast.Item3)
        return S.ExprEquation(lhs, expr, luid)
    if ast.IsSetSensorEquation:
        return S.SetSensorEquation(pathIdentifierOfAst(ast.Item1), expressionOfAst(ast.Item2))
    # def_by_case = automaton or activate
    return defByCaseOfAst(ast.Item, True)


def defByCaseOfAst(ast, is_equation=False):
    if ast.IsDAutomaton:  # of Lhs option * StateMachine * SourcePosition.t
        if lhs := getValueOf(ast.Item1):
            lhs = equationLhsOfAst(lhs)
        return stateMachineOfAst(lhs, ast.Item2, is_equation)

    if ast.DActivate:  # of Lhs option * Activate * SourcePosition.t
        if lhs := getValueOf(ast.Item1):
            lhs = equationLhsOfAst(lhs)
        if ast.Item2.IsActivateIf:
            return activateIfOfAst(lhs, ast.Item2, is_equation)
        # ast.Item2.IsActivateWhen
        return activateWhenOfAst(lhs, ast.Item2, is_equation)


# Activate
# ~~~~~~~~


# Activate if
def activateIfOfAst(lhs, ast, is_equation):
    # ActivateIf of luid option * IfActivation
    if name := getValueOf(ast.Item1):
        name = luidOfAst(name)
    activation = ifActivationOfAst(ast.Item2)
    return S.ActivateIf(activation, lhs, name, is_equation)


def ifActivationOfAst(ast):
    branches = [activationBranchOfAst(branch) for branch in ast.IfThenElif]
    else_branch = ifteBranchOfAst(ast.Else)
    branches.append(S.IfActivationBranch(None, else_branch))
    return S.IfActivation(branches)


def ifteBranchOfAst(ast):
    if ast.IsIfteDataDef:  # of ScopeDefinition
        data_def = scopeOfAst(ast.Item)
        return S.IfteDataDef(data_def)
    # ast.IfteBlock of IfActivation
    activation = ifActivationOfAst(ast.Item)
    return S.IfteIfActivation(activation)


def activationBranchOfAst(ast):
    expr = exprOrRawOfAst(ast.Item1)
    ifte_branch = ifteBranchOfAst(ast.Item2)
    return S.IfActivationBranch(expr, ifte_branch)


# Activate when
def activateWhenOfAst(lhs, ast, is_equation):
    # ActivateWhen of luid option * WhenActivation
    if name := getValueOf(ast.Item1):
        name = luidOfAst(name)
    condition = exprOrRawOfAst(ast.Item2.AWExpr)
    branches = [activateWhenBranchOfAst(branch) for branch in ast.Item2.AWMatches]
    return S.ActivateWhen(condition, branches, lhs, name, is_equation)


def activateWhenBranchOfAst(ast):
    # ast: PatternOrRaw * ScopeDefinition
    pattern = patternOrRawOfAst(ast.Item1)
    data_def = scopeOfAst(ast.Item2)
    return S.ActivateWhenBranch(pattern, data_def)


# State-machine
# ~~~~~~~~~~~~~


def stateMachineOfAst(lhs, ast, is_equation):
    if name := getValueOf(ast.Item1):
        name = luidOfAst(name)
    items = [stateMachineItemOfAst(item) for item in ast.Item2]
    machine = S.StateMachine(lhs, items, name, is_equation)
    return machine


def stateRefOfAst(ast):
    if ast.IsStateRefId:
        return S.StateRef(id=identifierOfAst(ast.Item))
    return S.StateRef(lunum=lunumOfAst(ast.Item))


def stateMachineItemOfAst(ast):
    if ast.IsStateItem:
        ast = ast.Item
        if id := getValueOf(ast.StateId):
            id = identifierOfAst(id)
        if lunum := getValueOf(ast.StateLunum):
            lunum = lunumOfAst(lunum)
        weak = [S.Transition(arrowOfAst(arrow)) for arrow in ast.UntilTransitions]
        strong = [S.Transition(arrowOfAst(arrow)) for arrow in ast.UnlessTransitions]
        is_initial = ast.StateIsInitial
        sections = stateBodyOfAst(ast.StateBody)
        pragmas = getPragmas(ast.StatePragmas)
        state = S.State(id, lunum, strong, sections, weak, is_initial, pragmas)
        return state

    # Transition declaration
    ast = ast.Item
    source = stateRefOfAst(ast.TSource)
    is_strong = ast.TStrong
    spec = arrowSpecOfAst(ast.TArrow)
    pragmas = getPragmas(ast.TPragmas)

    return S.TransitionDecl(spec["prio"], S.Transition(spec["arrow"], pragmas), is_strong, source)


def arrowOfAst(ast):
    spec = arrowSpecOfAst(ast)
    return spec["arrow"]


def arrowSpecOfAst(ast):
    prio = None if ast.APrio == "-1" else S.Literal(ast.APrio)

    if guard := getValueOf(ast.AGuard):
        guard = exprOrRawOfAst(guard)
    action = scopeOfAst(ast.AAction)

    arrow_target = None
    arrow_fork = None

    if fork := getValueOf(ast.AFork):  # AFork: Fork option
        if fork.IsAForkTree:
            # AForkTree of Arrow * Arrow list * Arrow option
            #  if guarded {{elsif guarded}} [else guarded]
            if_arrow = arrowOfAst(fork.Item1)
            elsif_arrows = [arrowOfAst(item) for item in fork.Item2]
            if else_arrow := getValueOf(fork.Item3):
                else_arrow = arrowOfAst(else_arrow)
            arrow_fork = S.ForkTree(if_arrow, elsif_arrows, else_arrow)

        else:
            # AForkPrio of Arrow list
            forks = [forkWithPrioFromAst(arrow) for arrow in fork.Item]
            arrow_fork = S.ForkPriorityList(forks)

    else:  # state
        target = getValueOf(ast.ATarget)
        assert target is not None
        target = stateRefOfAst(target)
        is_resume = ast.AIsResume
        arrow_target = S.Target(target, is_resume)

    # For Fork with priority: priority if guarded_arrow | priority else arrow
    is_if = ast.AIf

    return {
        "arrow": S.Arrow(guard, action, arrow_target, arrow_fork),
        "is_if": is_if,
        "prio": prio,
    }


def forkWithPrioFromAst(ast):
    arrow_spec = arrowSpecOfAst(ast)
    return S.ForkWithPriority(arrow_spec["prio"], arrow_spec["arrow"], arrow_spec["is_if"])


def stateBodyOfAst(ast):
    # StateBody : ScopeDefinition
    # but ScadeDefinition is as SDSections
    sections = [scopeSectionOfAst(section) for section in ast.Item1]
    return sections


# Diagram
# ---------------------------------------------------------------


def diagramObjectOfAst(ast):
    if lunum := getValueOf(ast.ObjLunum):
        lunum = lunumOfAst(lunum)
    if luid := getValueOf(ast.ObjLuid):
        luid = luidOfAst(luid)

    locals = [diagramObjectOfAst(obj) for obj in ast.ObjLocals]
    pragmas = getPragmas(ast.ObjPragmas)

    description = ast.ObjDescription

    if description.IsBExpr:  # ExprOrRaw
        expr = exprOrRawOfAst(description.Item)
        return S.ExprBlock(expr, lunum, luid, locals, pragmas)

    if description.IsBDef:  # Lhs * SourcePosition.t
        lhs = equationLhsOfAst(description.Item1)
        return S.DefBlock(lhs, lunum, luid, locals, pragmas)

    if description.IsBRawDef:  # Raw.t
        protected = protectedItemOfAst(description.Item)
        return S.DefBlock(protected, lunum, luid, locals, pragmas)

    if description.IsBBlock:  # OperatorBlock * SourcePosition.t
        op_block = operatorBlockOfAst(description.Item1)
        return S.Block(op_block, lunum=lunum, luid=luid, locals=locals, pragmas=pragmas)

    if description.IsBWire:  # Connection * Connection list
        source = connectionOfAst(description.Item1)
        targets = [connectionOfAst(conn) for conn in description.Item2]
        return S.Wire(source, targets, lunum, locals, pragmas)

    if description.IsBGroup:  # GroupOperation * SourcePosition.t
        ast_op = description.Item1
        if ast_op.IsGByName:
            operation = S.GroupOperation.ByName
        elif ast_op.IsGByPos:
            operation = S.GroupOperation.ByPos
        elif ast_op.IsGNoOp:
            operation = S.GroupOperation.NoOp
        else:  # IsGNorm
            operation = S.GroupOperation.Normalize

        return S.Bar(operation, lunum, locals, pragmas)

    if description.IsBDefByCase:
        def_by_case = defByCaseOfAst(description.Item)
        if isinstance(def_by_case, S.StateMachine):
            return S.StateMachineBlock(def_by_case, locals, pragmas)
        elif isinstance(def_by_case, S.ActivateIf):
            return S.ActivateIfBlock(def_by_case, locals, pragmas)
        else:
            return S.ActivateWhenBlock(def_by_case, locals, pragmas)

    if description.IsBScopeSection:  # ScopeSection
        section = scopeSectionOfAst(description.Item)
        return S.SectionBlock(section, locals, pragmas)

    # TODO test BSensorLhs, ...


def connectionOfAst(ast):
    if ast.IsConnEmpty:
        return S.Connection()

    # ConnPort of Port * GroupAdaptation option
    port = portOfAst(ast.Item1)
    if adaptation := getValueOf(ast.Item2):
        adaptation = groupAdaptationOfAst(adaptation)
    return S.Connection(port, adaptation)


def operatorBlockOfAst(ast):
    called = ast.OBCalled
    if called.IsCallOperator:  # Operator
        op_block = operatorBaseOfAst(called.Item)
        if isinstance(op_block, S.PathIdOpCall):
            cast(S.PathIdOpCall, op_block).pragmas.extend(getPragmas(ast.OBPragmas))
    else:  # CallOperatorExpr of OperatorExprWithSP
        op_block = operatorExprWithSPOfAst(called.Item)
    return op_block


# Scope & sections
# ~~~~~~~~~~~~~~~~
def scopeSectionOfAst(ast):
    if ast.IsSEmission:  # EmissionBody list * SourcePosition.t
        emissions = [emissionBodyOfAst(emit) for emit in ast.Item1]
        section = S.EmitSection(emissions)
        return section

    if ast.IsSAssume:  # VerifExpr list * SourcePosition.t
        hypotheses = [
            S.FormalProperty(luidOfAst(prop.VTag), expressionOfAst(prop.VExpr))
            for prop in ast.Item1
        ]
        section = S.AssumeSection(hypotheses)
        return section

    if ast.IsSAssert:  # VerifExpr list * SourcePosition.t
        assertions = [
            S.FormalProperty(luidOfAst(prop.VTag), expressionOfAst(prop.VExpr))
            for prop in ast.Item1
        ]
        section = S.AssertSection(assertions)
        return section

    if ast.IsSGuarantee:  # VerifExpr list * SourcePosition.t
        guarantees = [
            S.FormalProperty(luidOfAst(prop.VTag), expressionOfAst(prop.VExpr))
            for prop in ast.Item1
        ]
        section = S.GuaranteeSection(guarantees)
        return section

    if ast.IsSVarList:  # VarOrRaw list
        var_decls = [varDeclOfAst(var) for var in ast.Item]
        section = S.VarSection(var_decls)
        return section

    if ast.IsSLet:  # SourcePosition.t * Equation list * SourcePosition.t
        equations = [equationOfAst(eq) for eq in ast.Item2]
        section = S.LetSection(equations)
        return section

    if ast.IsSDiagram:  # Diagram
        objects = [diagramObjectOfAst(obj) for obj in ast.Item.DObjects]
        section = S.Diagram(objects)
        return section

    if ast.IsSRaw:  # Raw.t
        markup = getMarkup(ast.Item)
        content = getProtectedString(ast.Item)
        if markup == "text":
            origin = Parser.get_source().name
            swan = SwanString(f"{content}", origin)
            section = Parser.get_current_parser().scope_section(swan)
            section.is_text = True
            return section
        return S.ProtectedSection(getProtectedString(ast.Item))


def scopeOfAst(ast):
    if ast.IsSDEmpty:
        return None

    if ast.IsSDEquation:
        return equationOfAst(ast.Item)

    if ast.IsSDSections:
        sections = [scopeSectionOfAst(section) for section in ast.Item1]
        # diagram context
        pragmas = getPragmas(ast.Item3)
        scope = S.Scope(sections, pragmas)
        return scope


def operatorOfAst(ast):
    (
        inline,
        kind,
        name,
        inputs,
        outputs,
        sizes,
        constraints,
        specialization,
        pragmas,
    ) = signatureElementsOfAst(ast)

    def delayed_body(owner: S.SwanItem):
        if body := scopeOfAst(ast.OpBody):
            # body can be None
            body.owner = owner
        return body

    return S.Operator(
        id=name,
        is_inlined=inline,
        is_node=kind,
        inputs=inputs,
        outputs=outputs,
        body=delayed_body,
        sizes=sizes,
        constraints=constraints,
        specialization=specialization,
        pragmas=pragmas,
    )


def harnessOfAst(ast):
    name = stringOfStringWithSP(ast.HId)

    def delayed_body(owner: S.SwanItem):
        if body := scopeOfAst(ast.HBody):
            # body can be None
            body.owner = owner
        return body

    pragmas = getPragmas(ast.HPragmas)
    return S.TestHarness(
        id=name,  # path_id
        body=delayed_body,
        pragmas=pragmas,
    )


# Declaration factory
# ===================


def declarationOfAst(ast):
    """Build a ansys.scadeone.swan construct from an F# ast item

    Parameters
    ----------
    ast : F# object
        Object representing a declaration

    Returns
    -------
    GlobalDeclaration
        GlobalDeclaration derived object

    Raises
    ------
    ScadeOneException
        raise an exception when an invalid object is given
    """
    if ast.IsDConst:
        decls = [constDecl(item) for item in ast.Item1]
        return S.ConstDeclarations(decls)

    if ast.IsDGroup:
        decls = [groupDecl(item) for item in ast.Item1]
        return S.GroupDeclarations(decls)

    if ast.IsDOperator:
        return operatorOfAst(ast.Item)

    if ast.IsDSignature:
        return signatureOfAst(ast.Item)

    if ast.IsDSensor:
        decls = [sensorDecl(item) for item in ast.Item1]
        return S.SensorDeclarations(decls)

    if ast.IsDType:
        decls = [typeDecl(item) for item in ast.Item1]
        return S.TypeDeclarations(decls)

    if ast.IsDUse:
        return useDecl(ast.Item1)

    if ast.IsDRaw:
        markup = getMarkup(ast.Item)
        content = getProtectedString(ast.Item)
        if markup in ("text", "signature"):
            # - text denotes a *textual* operator declaration
            # (full operator or interface) in a module body.
            # It must be syntactically correct (else syntax_text apply)
            # - signature denotes a *textual* signature declaration in an interface:
            # It can either syntactically correct or not
            origin = Parser.get_source().name
            swan = SwanString(content, origin)
            if decl := Parser.get_current_parser().operator_decl(swan):
                decl.is_text = True
                return decl
            if markup == "text":
                raise ScadeOneException(f"invalid text operator declaration: {content}")
            # syntactically incorrect signature, default behavior

        # other protected: const, type, group, sensor, signature
        return S.ProtectedDecl(markup, content)

    if ast.IsDTestHarness:
        return harnessOfAst(ast.Item)

    raise ScadeOneException(f"unexpected ast class: {type(ast)}")


def allDeclsOfAst(ast):
    use_list = []
    decl_list = []
    for decl in ast.MDecls:
        py_obj = declarationOfAst(decl)
        if isinstance(py_obj, S.UseDirective):
            use_list.append(py_obj)
        else:
            decl_list.append(py_obj)
    return (use_list, decl_list)


def pathIdOfString(name: str) -> S.PathIdentifier:
    """Create a path identifier from a string

    Parameters
    ----------
    name : str
        Path name with '-' separating namespaces and module/interface
        name

    Returns
    -------
    S.PathIdentifier
         PathIdentifier object from name
    """
    if S.PathIdentifier.is_valid_file_path(name):
        id_list = [S.Identifier(id) for id in name.split("-")]
        return S.PathIdentifier(id_list)
    if S.PathIdentifier.is_valid_path(name):
        id_list = [S.Identifier(id.strip()) for id in name.split("::")]
        return S.PathIdentifier(id_list)
    return S.PathIdentifier(name)


def moduleOfAst(name: str, ast):
    path_id = pathIdOfString(name)
    (use_list, decl_list) = allDeclsOfAst(ast)
    body = S.ModuleBody(path_id, use_list, decl_list)
    return body


def interfaceOfAst(name, ast):
    path_id = pathIdOfString(name)
    (use_list, decl_list) = allDeclsOfAst(ast)
    interface = S.ModuleInterface(path_id, use_list, decl_list)
    return interface



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import json
from pathlib import Path
from typing import List, Optional, Union, TypedDict

from ansys.scadeone.core.common.exception import ScadeOneException
from ansys.scadeone.core.common.storage import (
    JobFile,
    ProjectFile,
    ProjectStorage,
    SwanFile,
)
from ansys.scadeone.core.common.versioning import FormatVersions
from ansys.scadeone.core.interfaces import IProject, IScadeOne
from ansys.scadeone.core.job import Job
from ansys.scadeone.core.svc.swan_creator.job_creator import JobFactory
from ansys.scadeone.core.model.model import Model
from ansys.scadeone.core.svc.swan_creator.project_creator import ProjectCreator
from ansys.scadeone.core.svc.swan_printer import swan_to_str
from ansys.scadeone.core.common.logger import LOGGER
from ansys.scadeone.core.swan.modules import Module


class _ProjectResource(TypedDict):
    """Project resource structure."""

    Kind: str
    Path: str
    Key: str


class _ProjectData(TypedDict):
    """Project data structure."""

    Version: str
    Name: str
    Dependencies: list[str]
    Resources: list[_ProjectResource]


class Project(IProject, ProjectCreator):
    """This class is the entry point of a project."""

    def __init__(self, app: IScadeOne, project: ProjectStorage) -> None:
        """Initialize a project file.

        Parameters
        ----------
        app : ScadeOne
            Application object.
        """
        self._app = app
        self._storage = project
        self._dependencies = None
        self._jobs = {}
        self._data = None
        self.is_new = False

    def save(self) -> None:
        """Save the project, that is all its modules.

        Two cases:

        - new project, saving to the `.sproj` location
        - loaded project, saving and replacing old files
        """

        if not self.is_new:
            self.model.load_all_modules()

        for module in self.model.modules:
            try:
                module_src = Path(module.source)
                module_src.parent.mkdir(parents=True, exist_ok=True)

                with module_src.open("w", newline="") as fd:
                    fd.write(swan_to_str(module))

                LOGGER.info(f"Saved: {module.source}")

            except Exception as e:
                raise ScadeOneException(f"Failed to save module {module.name.as_string}: {e}")

        if self.is_new:
            # Currently, no feature involves sproj edition
            # (dependencies, resources, project renaming...)
            self.build_sproj()

    def build_sproj(self) -> None:
        """Init SPROJ starting content."""

        Path(self._storage.path).parent.mkdir(parents=True, exist_ok=True)

        json_data = {
            "Version": FormatVersions.version("sproj"),
            "Name": self._storage.path.stem,
            "Dependencies": [],
            "Resources": [],
        }

        self._storage.set_content(json.dumps(json_data, indent=2))
        LOGGER.info(f"Created: {self._storage.source}")

    def check_exists(self, parent_path: Union[Path, str]) -> bool:
        """Return true if any of the project modules or sproj already exists
        attention, dependencies may not be checked."""

        if Path(self._storage.source).exists():
            return True
        for module in self.model.modules:
            file_path = parent_path / Path(module.source)
            if file_path.exists():
                return True
        return False

    def load_jobs(self) -> List[Job]:
        """(Re)load and return all the jobs of a project."""
        self._jobs = {}
        if self.directory:
            job_files = [
                JobFile(job) for job in self.directory.glob("jobs/**/*") if job.name == ".sjob"
            ]
            for job in job_files:
                job.load()
                job_json = job.json
                job_name = job_json["Properties"]["Name"]
                typed_job = JobFactory.create_job(job, self)
                if typed_job:
                    self._jobs[job_name] = typed_job
        return self.jobs

    def get_job(self, name: str) -> Union[Job, None]:
        """Get a job from its name."""
        return self._jobs.get(name)

    @property
    def jobs(self) -> List[Job]:
        """Return job files of the project."""
        return list(self._jobs.values())

    @property
    def app(self) -> IScadeOne:
        """Access to current Scade One application."""
        return self._app

    @property
    def model(self) -> Model:
        return self.app.model

    @property
    def modules(self) -> List[Module]:
        return self.model.modules

    @property
    def storage(self) -> ProjectStorage:
        """Project storage."""
        return self._storage

    @property
    def data(self) -> _ProjectData:
        """Project JSON data."""
        if self._data is None:
            self._data = self.storage.load().json
        return self._data

    @property
    def directory(self) -> Optional[Path]:
        """Project directory: Path if storage is a file, else None."""
        if isinstance(self.storage, ProjectFile):
            return Path(self.storage.path.parent.as_posix())
        return None

    def _get_swan_sources(self) -> List[SwanFile]:
        """Return Swan files of project.

        Returns
        -------
        List[SwanFile]
            List of SwanFile objects.
        """
        if self.directory is None:
            return []
        # glob uses Unix-style. Cannot have a fancy re, so need to check
        sources = [
            SwanFile(swan)
            for swan in self.directory.glob("assets/*.*")
            if swan.suffix in (".swan", ".swani", ".swant")
        ]
        return sources

    def swan_sources(self, all=False) -> List[SwanFile]:
        """Return all Swan sources from project.

        If all is True, include also sources from project dependencies.

        Returns
        -------
        list[SwanFile]
            List of all SwanFile objects.
        """
        sources = self._get_swan_sources()
        if all is False:
            return sources
        for lib in self.dependencies(all=True):
            sources.extend(lib.swan_sources())
        return sources

    def _get_dependencies(self) -> List["Project"]:
        """Projects directly referenced as dependencies.

        Returns
        -------
        list[Project]
            List of referenced projects.

        Raises
        ------
        ScadeOneException
            Raise exception if a project file does not exist.
        """
        if self._dependencies is not None:
            return self._dependencies
        if self.directory is None:
            return []

        def check_path(path: str) -> Path:
            s_path = self.app.subst_in_path(path).replace("\\", "/")
            p = Path(s_path)
            if not p.is_absolute():
                p = self.directory / p
            if p.exists():
                return p
            raise ScadeOneException(f"no such file: {path}")

        paths = [check_path(d) for d in self.data["Dependencies"]]
        self._dependencies = [Project(self._app, ProjectFile(p)) for p in paths]
        return self._dependencies

    def dependencies(self, all=False) -> List["Project"]:
        """Project dependencies as list of Projects.

        If all is True, include recursively dependencies of dependencies.

        A dependency occurs only once.
        """
        dependencies = self._get_dependencies()
        if not all:
            return dependencies

        # compute recursively all dependencies
        # One a project is visited, it is marked as visited
        # As returned Projects may be different objects project.dependencies() calls
        # one discriminates using the project source string.
        visited = {}

        for project in dependencies:
            source = project.storage.source
            if source in visited:
                continue

            visited[source] = project
            project_deps = project.dependencies(all=True)
            for sub_project in project_deps:
                sub_source = sub_project.storage.source
                if sub_source in visited:
                    continue
                visited[sub_source] = sub_project

        return list(visited.values())



# Copyright (C) 2024 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from enum import Enum, auto
from functools import cached_property
import json
import re
from typing import TYPE_CHECKING, List, Optional, Union, cast

from lark import Lark, Transformer

if TYPE_CHECKING:
    from ansys.scadeone.core.swan import Lunum


class PragmaBase:
    """Base class for objects with pragmas."""

    def __init__(self, pragmas: Optional[List["Pragma"]] = None) -> None:
        self._pragmas = pragmas if pragmas else []

    @property
    def pragmas(self) -> List["Pragma"]:
        """List of pragmas."""
        return self._pragmas

    def pragmas_str(self) -> str:
        """Return a string with all pragmas."""
        pragmas = " ".join(str(p) for p in self.pragmas)
        return pragmas


class Pragma:
    """Pragma structure."""

    def __init__(self, pragma: str) -> None:
        self._pragma = pragma

    @property
    def pragma(self) -> str:
        """Return full pragma string."""
        return self._pragma

    @cached_property
    def diagram(self) -> "DiagramPragma":
        """Return the pragma diagram."""
        parser = PragmaParser()
        parser.parse(self._pragma)
        return parser.diagram

    @staticmethod
    def filter(pragmas: List["Pragma"], key: str, with_key: bool = True) -> List["Pragma"]:
        """Filters a list of pragmas with/without a given key.

        Parameters
        ----------
        pragmas : List[Pragma]
            List of pragmas.
        key : str
            Key to filter.
        with_key : bool, optional
            If True, return pragmas with the given key, otherwise without the key, by default True.

        Returns
        -------
        List[Pragma]
            List of pragmas with the given key.
        """
        res = []
        for p in pragmas:
            m = PragmaParser.extract(p.pragma)
            if m and (with_key == (m[0] == key)):
                res.append(p)
        return res

    def __str__(self) -> str:
        return self._pragma


class DiagramPragma:
    """The diagram information of a graphical object."""

    def __init__(self) -> None:
        self._coordinates = None
        self._size = None
        self._direction = None
        self._orientation = None
        self._wire_info = None
        self._arrow_info = None

    @property
    def coordinates(self) -> "Coordinates":
        """Return the diagram coordinates."""
        return self._coordinates

    @property
    def size(self) -> "Size":
        """Return the diagram size."""
        return self._size

    @property
    def direction(self) -> "Direction":
        """Return the diagram direction."""
        return self._direction

    @property
    def orientation(self) -> "Orientation":
        """Return the diagram orientation."""
        return self._orientation

    @property
    def wire_info(self) -> "PathInfo":
        """Return the diagram wire info."""
        return self._wire_info

    @property
    def arrow_info(self) -> "PathInfo":
        """Return the diagram arrow info."""
        return self._arrow_info

    def __str__(self):
        params = []
        if self._coordinates:
            params.append(f'"xy":"{self._coordinates}"')
        if self._size:
            params.append(f'"wh":"{self._size}"')
        if self._direction:
            params.append(f'"dir":"{self._direction}"')
        if self._orientation:
            params.append(f'"orient":"{self._orientation}"')
        if self._wire_info:
            params.append(f'"wp":"{self._wire_info}"')
        if self._arrow_info:
            params.append(f'"tp":"{self._arrow_info}"')
        if len(params) == 0:
            return ""
        return f"{{{','.join(params)}}}"


def _create_parser(grammar: str, start: str, transformer: Transformer) -> Lark:
    """Create a Lark parser with the given grammar, start and transformer. LALR parser is used."""
    return Lark(grammar, start=start, parser="lalr", transformer=transformer)


class PragmaParser:
    """Parser for pragma.

    This is the list of supported pragmas:
    - diagram
    """

    PragmaRE = re.compile(r"#pragma\s+(?P<key>\w+)\s(?P<val>.*)#end", re.DOTALL)
    _instance = None

    def __init__(self) -> None:
        self._diagram = None

    def __new__(cls, *args, **kwargs) -> "PragmaParser":
        if not cls._instance:
            cls._instance = super(PragmaParser, cls).__new__(cls)
        return cls._instance

    @property
    def diagram(self) -> "DiagramPragma":
        """Return the diagram pragma."""
        return self._diagram

    def parse(self, pragma: str) -> None:
        """Parse pragma.
        This is the list of supported pragmas:
        - diagram

        Parameters
        ----------
            pragma : str
                Pragma string defined in SO-SRS-001 V2.1, section 1.2.5, [S1-203]
        """
        if not pragma:
            return
        pragma_tuple = self.extract(pragma)
        if not pragma_tuple:
            return
        key, value = pragma_tuple
        if key == "diagram":
            self._diagram = DiagramPragmaParser().parse(value)
            if self._diagram and not self._diagram.direction:
                self._diagram._direction = Direction(DirectionType.NORTH_EAST)

    @staticmethod
    def extract(pragma: str) -> Union[tuple[str, str], None]:
        """Extract pragma information as a tuple
        if pragma is valid, namely: #pragma key value#end.

        Returns
        -------
            tuple | None
                The Tuple (*pragma name*, *pragma value*) if pragma is valid, None else.
        """
        m = PragmaParser.PragmaRE.match(pragma)
        if not m:
            return None
        return m["key"], m["val"].strip()


class DiagramPragmaParser:
    """Parser for diagram pragma.
    The different parsers implement SO-SRS-001 V2.1, section 5.5: Graphical Information
    """

    _instance = None

    def __init__(self) -> None:
        self._coordinates_parser = self._create_coordinate_parser()
        self._size_parser = self._create_size_parser()
        self._direction_parser = self._create_direction_parser()
        self._orientation_parser = self._create_orientation_parser()
        self._path_info_parser = self._create_path_info_parser()

    def __new__(cls, *args, **kwargs) -> "DiagramPragmaParser":
        if not cls._instance:
            cls._instance = super(DiagramPragmaParser, cls).__new__(cls)
        return cls._instance

    @staticmethod
    def _create_coordinate_parser() -> Lark:
        """Create the parser for coordinates.
        Coordinates are defined as "[Hh]x;[Vv]y" where:

        - H,V: Absolute position
        - h,v: Relative position
        - x,y: Coordinate values

        Returns
        -------
        Lark
            A coordinates' parser.
        """
        grammar = r"""
                    coordinates: x ";" y
                    x: horizontal number
                    y: vertical number
                    horizontal: absolute_horizontal | relative_horizontal
                    vertical: absolute_vertical | relative_vertical
                    absolute_horizontal: "H"
                    relative_horizontal: "h"
                    absolute_vertical: "V"
                    relative_vertical: "v"
                    number: /-?\d+/
                    %import common.WS
                    %ignore WS
                    """
        return _create_parser(grammar, start="coordinates", transformer=CoordinateTransformer())

    @staticmethod
    def _create_size_parser() -> Lark:
        """Create the parser for size.
        Size is defined as "width;height" where:

        - width: Width value
        - height: Height value

        Returns
        -------
        Lark
            A size parser.
        """
        grammar = r"""
                    size: width ";" height
                    width: number
                    height: number
                    number: /-?\d+/
                    %import common.WS
                    %ignore WS
                    """
        return _create_parser(grammar, start="size", transformer=SizeTransformer())

    @staticmethod
    def _create_direction_parser() -> Lark:
        """Create the parser for a direction.
        Direction is defined as "ne|nw|es|en|se|sw|ws|wn" where:

        - ne: North east
        - nw: North west
        - es: East south
        - en: East north
        - se: South east
        - sw: South west
        - ws: West south
        - wn: West north

        Returns
        -------
        Lark
            A direction parser.
        """
        grammar = r"""
                    direction: north_east | north_west | east_south | east_north | south_east | south_west | west_south | west_north
                    north_east: "ne"
                    north_west: "nw"
                    east_south: "es"
                    east_north: "en"
                    south_east: "se"
                    south_west: "sw"
                    west_south: "ws"
                    west_north: "wn"
                    %import common.WS
                    %ignore WS
                    """
        return _create_parser(grammar, start="direction", transformer=DirectionTransformer())

    @staticmethod
    def _create_orientation_parser() -> Lark:
        """Create the parser for orientation.
        Orientation is defined as "H|V" where:

        - H: Horizontal
        - V: Vertical

        Returns
        -------
        Lark
            An orientation parser.
        """
        grammar = r"""
                    orientation: horizontal | vertical
                    horizontal: "H"
                    vertical: "V"
                    %import common.WS
                    %ignore WS
                    """
        return _create_parser(grammar, start="orientation", transformer=OrientationTransformer())

    @staticmethod
    def _create_path_info_parser() -> Lark:
        """Create the parser for path info.
        Path info is defined as "w_anchor path" where:

        - w_anchor: LUNUM
                    | COORD '|' LUNUM
                    | coordinates '|' LUNUM
                    | coordinates
        - path: moves w_anchor
                | moves branch
        - moves: /* empty */
                | moves move
        - move: COORD
                | COORD '|' coordinates
                | coordinates '|' coordinates
                | coordinates
        - coordinates: x ';' y
        - branch: '[' path_list ']'
        - path_list: path_list ',' path

        Returns
        -------
        Lark
            A path info parser.
        """
        grammar = r"""
                    path_info: w_anchor path
                    w_anchor: LUNUM
                        | (x | y | coordinates) "|" LUNUM
                        | coordinates
                    path: move* w_anchor | move* branch
                    coordinates: x ";" y
                    x: absolute_horizontal | relative_horizontal SIGNED_INT
                    y: absolute_vertical | relative_vertical SIGNED_INT
                    absolute_horizontal: "H"
                    relative_horizontal: "h"
                    absolute_vertical: "V"
                    relative_vertical: "v"
                    move: (x | y | coordinates)
                        | fork_coordinates
                    fork_coordinates: (x | y | coordinates) "|" coordinates
                    branch: "[" path_list "]"
                    path_list: (path ",")* path
                    LUNUM: "#" DIGIT10+
                    DIGIT10: /[0-9]/
                    %import common.WS
                    %import common.SIGNED_INT
                    %ignore WS
                    """
        return _create_parser(grammar, start="path_info", transformer=PathInfoTransformer())

    def parse(self, params: str) -> Union["DiagramPragma", None]:
        """Parse pragma diagram.

        Parameters
        ----------
            params : str
                Pragma diagram parameters.
                This is a string that contains a JSON expression with the following properties:

                - "xy": Coordinates
                - "wh": Size
                - "dir": Direction
                - "orient": Orientation
                - "wp": Wire path
                - "tp": Arrow path

            Each property's value is parsed by the corresponding parser.

        Returns
        -------
            DiagramPragma | None
                The parsed pragma diagram or None if params is empty.
        """
        if not params:
            return None
        params = json.loads(params)
        if not isinstance(params, dict):
            from ansys.scadeone.core import ScadeOneException

            raise ScadeOneException(f"Pragma diagram must be a dictionary: {params}")
        pragma_diag = DiagramPragma()
        if "xy" in params:
            pragma_diag._coordinates = cast(
                Coordinates, self._coordinates_parser.parse(params["xy"])
            )
        if "wh" in params:
            pragma_diag._size = cast(Size, self._size_parser.parse(params["wh"]))
        if "dir" in params:
            pragma_diag._direction = cast(Direction, self._direction_parser.parse(params["dir"]))
        if "orient" in params:
            pragma_diag._orientation = cast(
                Orientation, self._orientation_parser.parse(params["orient"])
            )
        if "wp" in params:
            pragma_diag._wire_info = cast(PathInfo, self._path_info_parser.parse(params["wp"]))
        if "tp" in params:
            pragma_diag._arrow_info = cast(PathInfo, self._path_info_parser.parse(params["tp"]))
        return pragma_diag


class CoordinateTransformer(Transformer):
    """Coordinate transformer.
    Transform the parser tree into Coordinates."""

    @staticmethod
    def coordinates(items) -> "Coordinates":
        """Return the coordinates.

        Parameters
        ----------
        items : list
            List of items. This list only contains two items:
            - The x coordinate
            - The y coordinate

        Returns
        -------
        Coordinates
            The object coordinates (diagram object, states or Active if/when blocks).
        """
        return Coordinates(items[0], items[1])

    @staticmethod
    def x(items) -> "Coordinate":
        """Return the *x* coordinate.

        Parameters
        ----------
        items : list
            List of items. This list only contains two items:
            - The position of the x coordinate
            - The value of the x coordinate

        Returns
        -------
        Coordinate
            The x coordinate.
        """
        return Coordinate(items[0], items[1])

    @staticmethod
    def y(items) -> "Coordinate":
        """Return the *y* coordinate.

        Parameters
        ----------
        items : list
            List of items. This list only contains two items:
            - The position of the y coordinate
            - The value of the y coordinate

        Returns
        -------
        Coordinate
            The y coordinate.
        """
        return Coordinate(items[0], items[1])

    @staticmethod
    def horizontal(items) -> "Position":
        """Return the horizontal position.

        Parameters
        ----------
        items : list
            List of items. This list only contains one item:
            - The horizontal position (absolute or relative)

        Returns
        -------
        Position
            The horizontal position (absolute or relative).
        """
        return items[0]

    @staticmethod
    def vertical(items) -> "Position":
        """Return the vertical position.

        Parameters
        ----------
        items : list
            List of items. This list only contains one item:
            - The vertical position (absolute or relative)

        Returns
        -------
        Position
            The vertical position (absolute or relative).
        """
        return items[0]

    @staticmethod
    def absolute_horizontal(items) -> "Position":
        """Return an absolute position.

        Parameters
        ----------
        items : list
            List of items. It's not used but needed for the parser.
            When the parser calls this method, it expects an absolute position for x.

        Returns
        -------
        Position
            Absolute x position.
        """
        return Position.ABSOLUTE

    @staticmethod
    def relative_horizontal(items) -> "Position":
        """Return a relative position.

        Parameters
        ----------
        items : list
            List of items. It's not used but needed for the parser.
            When the parser calls this method, it expects a relative position for x.

        Returns
        -------
        Position
            Relative x position.
        """
        return Position.RELATIVE

    @staticmethod
    def absolute_vertical(items) -> "Position":
        """Return an absolute position.

        Parameters
        ----------
        items : list
            List of items. It's not used but needed for the parser.
            When the parser calls this method, it expects an absolute position for y.

        Returns
        -------
        Position
            Absolute y position.
        """
        return Position.ABSOLUTE

    @staticmethod
    def relative_vertical(items) -> "Position":
        """Return a relative position.

        Parameters
        ----------
        items : list
            List of items. It's not used but needed for the parser.
            When the parser calls this method, it expects a relative position for y.

        Returns
        -------
        Position
            Relative y position.
        """
        return Position.RELATIVE

    @staticmethod
    def number(items) -> int:
        """Return the coordinate value.

        Parameters
        ----------
        items : list
            List of items. This list only contains one item:
            - The coordinate value

        Returns
        -------
        int
            The coordinate value.
        """
        return int(items[0].value)


class SizeTransformer(Transformer):
    """Size transformer.
    Transform the parser tree into Size."""

    @staticmethod
    def size(items) -> "Size":
        """Return the size.

        Parameters
        ----------
        items : list
            List of items. This list only contains two items:
            - The width
            - The height

        Returns
        -------
        Size
            The object size.
        """
        return Size(items[0], items[1])

    @staticmethod
    def width(items) -> int:
        """Return the width.

        Parameters
        ----------
        items : list
            List of items. This list only contains one item:
            - The width value

        Returns
        -------
        int
            The object width value.
        """
        return items[0]

    @staticmethod
    def height(items) -> int:
        """Return the height.

        Parameters
        ----------
        items : list
            List of items. This list only contains one item:
            - The height value

        Returns
        -------
        int
            The object height value.
        """
        return items[0]

    @staticmethod
    def number(items) -> int:
        """Return the size value.

        Parameters
        ----------
        items : list
            List of items. This list only contains one item:
            - The size value

        Returns
        -------
        int
            The object size value.
        """
        return int(items[0].value)


class DirectionTransformer(Transformer):
    """Direction transformer.
    Transforms the parser tree into Direction."""

    @staticmethod
    def direction(items) -> "Direction":
        """Return the direction.

        Parameters
        ----------
        items : list
            List of items. This list only contains one item:
            - The direction value

        Returns
        -------
        Direction
            The object direction.
        """
        return Direction(items[0])

    @staticmethod
    def north_east(items) -> "DirectionType":
        """Return the north-east direction.

        Parameters
        ----------
        items : list
            List of items. It's not used but needed for the parser.
            When the parser calls this method, it expects the north-east direction.

        Returns
        -------
        DirectionType
            The north-east direction
        """
        return DirectionType.NORTH_EAST

    @staticmethod
    def north_west(items) -> "DirectionType":
        """Return the north-west direction.

        Parameters
        ----------
        items : list
            List of items. It's not used but needed for the parser.
            When the parser calls this method, it expects the north-west direction.

        Returns
        -------
        DirectionType
            The north-west direction
        """
        return DirectionType.NORTH_WEST

    @staticmethod
    def east_south(items) -> "DirectionType":
        """Return the east-south direction.

        Parameters
        ----------
        items : list
            List of items. It's not used but needed for the parser.
            When the parser calls this method, it expects the east-south direction.

        Returns
        -------
        DirectionType
            The east-south direction
        """
        return DirectionType.EAST_SOUTH

    @staticmethod
    def east_north(items) -> "DirectionType":
        """Return the east-north direction.

        Parameters
        ----------
        items : list
            List of items. It's not used but needed for the parser.
            When the parser calls this method, it expects the east-north direction.

        Returns
        -------
        DirectionType
            The east-north direction
        """
        return DirectionType.EAST_NORTH

    @staticmethod
    def south_east(items) -> "DirectionType":
        """Return the south-east direction.

        Parameters
        ----------
        items : list
            List of items. It's not used but needed for the parser.
            When the parser calls this method, it expects the south-east direction.

        Returns
        -------
        DirectionType
            The south-east direction
        """
        return DirectionType.SOUTH_EAST

    @staticmethod
    def south_west(items) -> "DirectionType":
        """Return the south-west direction.

        Parameters
        ----------
        items : list
            List of items. It's not used but needed for the parser.
            When the parser calls this method, it expects the south-west direction.

        Returns
        -------
        DirectionType
            The south-west direction
        """
        return DirectionType.SOUTH_WEST

    @staticmethod
    def west_south(items) -> "DirectionType":
        """Return the west-south direction.

        Parameters
        ----------
        items : list
            List of items. It's not used but needed for the parser.
            When the parser calls this method, it expects the west-south direction.

        Returns
        -------
        DirectionType
            The west-south direction
        """
        return DirectionType.WEST_SOUTH

    @staticmethod
    def west_north(items) -> "DirectionType":
        """Return the west-north direction.

        Parameters
        ----------
        items : list
            List of items. It's not used but needed for the parser.
            When the parser calls this method, it expects the west-north direction.

        Returns
        -------
        DirectionType
            The west-north direction
        """
        return DirectionType.WEST_NORTH


class OrientationTransformer(Transformer):
    """Orientation transformer.
    Transforms the parser tree into Orientation."""

    @staticmethod
    def orientation(items) -> "Orientation":
        """Return the orientation.

        Parameters
        ----------
        items : list
            List of items. This list only contains one item:
            - The orientation value

        Returns
        -------
        Orientation
            The object orientation.
        """
        return Orientation(items[0])

    @staticmethod
    def horizontal(items) -> "OrientationType":
        """Return the horizontal orientation.

        Parameters
        ----------
        items : list
            List of items. It's not used but needed for the parser.
            When the parser calls this method, it expects the horizontal orientation.

        Returns
        -------
        OrientationType
            The horizontal orientation
        """
        return OrientationType.HORIZONTAL

    @staticmethod
    def vertical(items) -> "OrientationType":
        """Return the vertical orientation.

        Parameters
        ----------
        items : list
            List of items. It's not used but needed for the parser.
            When the parser calls this method, it expects the vertical orientation.

        Returns
        -------
        OrientationType
            The vertical orientation
        """
        return OrientationType.VERTICAL


class PathInfoTransformer(Transformer):
    """Transforms the parser tree into PathInfo."""

    @staticmethod
    def path_info(items) -> "PathInfo":
        """Return the path info.

        Parameters
        ----------
        items : list
            List of items. This list only contains two items:
            - The wire anchor
            - The path

        Returns
        -------
        PathInfo
            The object path info.
        """
        return PathInfo(items[0], items[1])

    @staticmethod
    def w_anchor(items) -> "WireAnchor":
        """Return the wire anchor.

        Parameters
        ----------
        items : list
            List of items. This list could contain two items:
            - The LUNUM
            - The coordinates

        Returns
        -------
        WireAnchor
            The object wire anchor.
        """
        from ansys.scadeone.core.swan import Lunum

        lunum = None
        coordinates = None
        for item in items:
            if isinstance(item, Lunum):
                lunum = item
            elif isinstance(item, tuple):
                if item[0] == "x":
                    coordinates = Coordinates(item[1], None)
                else:
                    coordinates = Coordinates(None, item[1])
            elif isinstance(item, Coordinates):
                coordinates = item
        return WireAnchor(lunum, coordinates)

    @staticmethod
    def path(items) -> "WirePath":
        """Return the wire path.

        Parameters
        ----------
        items : list
            List of items. This list could contain three items:
            - The moves
            - The wire anchor
            - The branch

        Returns
        -------
        WirePath
            The object wire path.
        """
        moves = None
        w_anchor = None
        branch = None
        for item in items:
            if isinstance(item, Move):
                if not moves:
                    moves = []
                moves.append(item)
            if isinstance(item, WireAnchor):
                w_anchor = item
            if isinstance(item, Branch):
                branch = item
        return WirePath(moves, w_anchor, branch)

    @staticmethod
    def coordinates(items) -> "Coordinates":
        """Return the coordinates.

        Parameters
        ----------
        items : list
            List of items. This list only contains two items:
            - The *x* coordinate
            - The *y* coordinate

        Returns
        -------
        Coordinates
            The object coordinates.
        """
        return Coordinates(items[0][1], items[1][1])

    @staticmethod
    def x(items) -> tuple[str, "Coordinate"]:
        """Return the *x* coordinate.

        Parameters
        ----------
        items : list
            List of items. This list only contains two items:
            - The *x* position
            - The *x* value

        Returns
        -------
        tuple
            The *x* coordinate. The tuple is used to identify the *x* or *y* coordinate.
        """
        return "x", Coordinate(items[0], items[1].value)

    @staticmethod
    def y(items) -> tuple[str, "Coordinate"]:
        """Return the *y* coordinate.

        Parameters
        ----------
        items : list
            List of items. This list only contains two items:
            - The *y* position
            - The *y* value

        Returns
        -------
        tuple
            The *y* coordinate. The tuple is used to identify the *x* or *y* coordinate.
        """
        return "y", Coordinate(items[0], items[1].value)

    @staticmethod
    def absolute_horizontal(items) -> "Position":
        """Return an absolute position.

        Parameters
        ----------
        items : list
            List of items. It's not used but needed for the parser.
            When the parser calls this method, it expects an absolute position for *x*.

        Returns
        -------
        Position
            *x* is in absolute position.
        """
        return Position.ABSOLUTE

    @staticmethod
    def relative_horizontal(items) -> "Position":
        """Return a relative position.

        Parameters
        ----------
        items : list
            List of items. It's not used but needed for the parser.
            When the parser calls this method, it expects a relative position for *x*.

        Returns
        -------
        Position
            *x* is in relative position.
        """
        return Position.RELATIVE

    @staticmethod
    def absolute_vertical(items) -> "Position":
        """Return an absolute position.

        Parameters
        ----------
        items : list
            List of items. It's not used but needed for the parser.
            When the parser calls this method, it expects an absolute position for *y*.

        Returns
        -------
        Position
            *y* is in absolute position.
        """
        return Position.ABSOLUTE

    @staticmethod
    def relative_vertical(items) -> "Position":
        """Return a relative position.

        Parameters
        ----------
        items : list
            List of items. It's not used but needed for the parser.
            When the parser calls this method, it expects a relative position for *y*.

        Returns
        -------
        Position
            *y* is in relative position.
        """
        return Position.RELATIVE

    @staticmethod
    def move(items) -> "Move":
        """Returns the move.

        move: (x | y | coordinates) '|' coordinates

        Parameters
        ----------
        items : list
            List of items. This list only contains an item:
            - tuple: The *x* or *y* coordinate
            - The coordinates
            - The coordinates with fork coordinates

        Returns
        -------
        Move
            The object move.
        """
        move_coordinates = None
        if isinstance(items[0], tuple):
            if items[0][0] == "x":
                coordinates = Coordinates(items[0][1], None)
            else:
                coordinates = Coordinates(None, items[0][1])
            move_coordinates = MoveCoordinates(coordinates)
        elif isinstance(items[0], Coordinates):
            move_coordinates = MoveCoordinates(items[0])
        elif isinstance(items[0], MoveCoordinates):
            move_coordinates = items[0]
        return Move(move_coordinates)

    @staticmethod
    def fork_coordinates(items) -> "MoveCoordinates":
        """Return the coordinates with fork coordinates.

        fork_coordinates: (x | y | coordinates) '|' coordinates

        Parameters
        ----------
        items : list
            List of items. This list only contains two items:
            - The coordinates
            - The fork coordinates
        """
        return MoveCoordinates(items[0], items[1])

    @staticmethod
    def branch(items) -> "Branch":
        """Return the branch.

        branch: '[' path_list ']'

        Parameters
        ----------
        items : list
            List of items. This list only contains one item:
            - The path list

        Returns
        -------
        Branch
            The object branch.
        """
        return Branch(items[0])

    @staticmethod
    def path_list(items) -> List["WirePath"]:
        """Return the path list.

        path_list: (path ',')* path

        Parameters
        ----------
        items : list
            List of items. This list contains the path list.

        Returns
        -------
        List[WirePath]
            The object path list.
        """
        return items

    @staticmethod
    def LUNUM(item) -> "Lunum":
        """Return the *LUNUM*.

        Parameters
        ----------
        item : str
            The *LUNUM* value.

        Returns
        -------
        Lunum
            The object *LUNUM*.
        """
        from ansys.scadeone.core.swan import Lunum

        return Lunum(item.value)

    @staticmethod
    def DIGIT10(item) -> str:
        """Return the *LUNUM* number.

        Parameters
        ----------
        item : str
            The *LUNUM* number.

        Returns
        -------
        str
            The object *LUNUM* number.
        """
        return item.value


class Coordinates:
    """*Coordinates* define a horizontal (*x*) and a vertical (*y*) position of a graphical object.

    *Coordinates* are specified as:

    .. code-block:: ebnf

        xy = COORD ";" COORD

    where:

    - :code:`COORD`: Coordinate *x* or *y* (see :py:class:`Coordinate`)

    *Coordinates* are used to define the position of the diagram object, states or Active if/when blocks.
    """

    def __init__(self, x: Optional["Coordinate"] = None, y: Optional["Coordinate"] = None) -> None:
        self._x = x
        self._y = y

    @property
    def x(self) -> Union["Coordinate", None]:
        """Return the *x* coordinate."""
        return self._x

    @property
    def y(self) -> Union["Coordinate", None]:
        """Return the *y* coordinate."""
        return self._y

    def __str__(self) -> str:
        coordinates = []
        if self._x:
            position_str = "H" if self._x.position == Position.ABSOLUTE else "h"
            coordinates.append(f"{position_str}{self._x.value}")
        if self._y:
            position_str = "V" if self._y.position == Position.ABSOLUTE else "v"
            coordinates.append(f"{position_str}{self._y.value}")
        if len(coordinates) == 0:
            return ""
        return ";".join(coordinates)


class Coordinate:
    """*Coordinate* defines a horizontal (*x*) or vertical (*y*) position of a graphical object.

    *Coordinate* is defined as:

    .. code-block:: ebnf

        xy = ("H" | "h")x ";" ("V" | "v")y

    - :code:`("H" | "h")` or :code:`("V" | "v")`: Absolute or relative position (see :py:class:`Position`)
    - :code:`x` or :code:`y`: Coordinate value

    The relative coordinate values are computed from:

    - the center of the parent block
    - wire and transition: the center of the source/target, or the previous coordinate
    """

    def __init__(self, position: "Position", value: int) -> None:
        self._position = position
        self._value = value

    @property
    def position(self) -> "Position":
        """Return the coordinate position (absolute or relative)."""
        return self._position

    @property
    def value(self) -> int:
        """Return the coordinate value."""
        return self._value

    def __str__(self):
        return f"{self._position.name}:{self._value}"


class Position(Enum):
    """Position of a coordinate.

    .. code-block:: ebnf

        xy = ("H" | "h")x ";" ("V" | "v")y

    - :code:`H`, :code:`V`: Absolute position
    - :code:`h`, :code:`v`: Relative position
    """

    #: Absolute position.
    ABSOLUTE = auto()

    #: Relative position.
    RELATIVE = auto()


class Size:
    """Size of a graphical object.

    Size is defined as

    .. code-block:: ebnf

        wh = width ";" height

    where:

    - :code:`width`: Width value
    - :code:`height`: Height value

    Size is used to define the size of the diagram object,
    states, active if/when blocks or automaton.
    """

    def __init__(self, width: int, height: int) -> None:
        self._width = width
        self._height = height

    @property
    def width(self) -> int:
        """Return the object width."""
        return self._width

    @property
    def height(self) -> int:
        """Return the object height."""
        return self._height

    def __str__(self):
        return f"{self._width};{self._height}"


class Direction:
    """The direction of the diagram object.

    *Direction* is defined as

    .. code-block:: ebnf

        dir = dir_val

    where :code:`dir_val` is defined in :py:class:`DirectionType`.

    *Direction* is used to define the direction of predefined operator and block with text (Expr, Def, Instance, Equation).
    """

    def __init__(self, value: "DirectionType") -> None:
        self._value = value

    @property
    def value(self) -> "DirectionType":
        """Return the direction value."""
        return self._value

    def __str__(self):
        return self._value.value


class DirectionType(Enum):
    """The direction value of a diagram object.

    The direction value is defined as

    .. code-block:: ebnf

        dir_val = ne|nw|es|en|se|sw|ws|wn

    where:

    - :code:`ne`: North-East
    - :code:`nw`: North-West
    - :code:`es`: East-South
    - :code:`en`: East-North
    - :code:`se`: South-East
    - :code:`sw`: South-West
    - :code:`ws`: West-South
    - :code:`wn`: West-North

    The direction value is read as: the first direction is at the top of the graphical object,
    and the second direction is the right side of the graphical object.

    The default direction is North-East:
        - West corresponding to inputs,
        - East corresponding to outputs.
    """

    #: North-East direction. Default direction (inputs are on the left, outputs are on the right).
    NORTH_EAST = "ne"

    #: North-West direction. From the default direction, a horizontal flip is applied,
    #: or a 180° rotation and a vertical flip are applied
    #: (inputs are on the right and outputs are on the left).
    NORTH_WEST = "nw"

    #: East-South direction. From the default direction, a 90° left rotation is applied
    #: (inputs are at the bottom and outputs are at the top).
    EAST_SOUTH = "es"

    #: East-North direction. From the default direction, a 90° left rotation and a horizontal flip are applied
    #: (inputs are at the bottom and outputs are at the top).
    EAST_NORTH = "en"

    #: South east direction. From the default direction, a vertical flip is applied
    #: (inputs are on the left and outputs are on the right).
    SOUTH_EAST = "se"

    #: South-West direction. From the default direction, a 180° rotation is applied
    #: (inputs are on the right and outputs are on the left).
    SOUTH_WEST = "sw"

    #: West-South direction. From the default direction, a 90° right rotation and a horizontal flip are applied
    #: (inputs are at the top and outputs are at the bottom).
    WEST_SOUTH = "ws"

    #: West-North direction. From the default direction a 90° right rotation is applied
    #: (inputs are at the top and outputs are at the bottom).
    WEST_NORTH = "wn"


class Orientation:
    """Text content orientation.

    *Orientation* is defined as

    .. code-block:: ebnf

        orient = orient_val

    where :code:`orient_val` is defined in :py:class:`OrientationType`.

    *Orientation* is used to define the text content orientation of the diagram object.
    """

    def __init__(self, value: "OrientationType") -> None:
        self._value = value

    @property
    def value(self) -> "OrientationType":
        """Return the orientation value."""
        return self._value

    def __str__(self):
        if self._value is None:
            return ""
        return self._value.value


class OrientationType(Enum):
    """Text content orientation value.

    The orientation value is defined as:

    .. code-block:: ebnf

        orient_val = "H"|"V"

    where:

    - :code:`H`: Horizontal orientation
    - :code:`V`: Vertical orientation
    """

    #: Horizontal orientation.
    HORIZONTAL = "H"

    #: Vertical orientation.
    VERTICAL = "V"


class PathInfo:
    """The wire or transition path between two objects.
    The path is defined by a list of moves.

    The wire and transition paths are defined as:

    - Wire

    .. code-block:: ebnf

        wp = path_info

    - Transition

    .. code-block:: ebnf

        tp = path_info

    where:

    .. code-block:: ebnf

        path_info = w_anchor path

    :code:`w_anchor` is defined in :py:class:`WireAnchor`.

    :code:`path` is defined in :py:class:`WirePath`.
    """

    def __init__(self, w_anchor: "WireAnchor", path: "WirePath") -> None:
        self._w_anchor = w_anchor
        self._path = path

    @property
    def w_anchor(self) -> "WireAnchor":
        """Return the wire anchor."""
        return self._w_anchor

    @property
    def path(self) -> "WirePath":
        """Return the path."""
        return self._path

    def __str__(self):
        return f"{self._w_anchor} {self._path}"


class WireAnchor:
    """*Wire anchor* is the starting or ending point of a path.

    *Wire anchor* is defined as:

    .. code-block:: ebnf

        w_anchor = LUNUM
            | COORD '|' LUNUM (* connection to a group-related block *)
            | coordinates '|' LUNUM (* coordinates of a starting/ending point of a transition for a state *)
            | coordinates (* unconnected point as a pair of COORD. *)

    where:

    - :code:`LUNUM`: The graphical object identifier (see :py:class:`Lunum`)
    - :code:`COORD`: Coordinate *x* or *y* (see :py:class:`Coordinate`)
    - :code:`coordinates`: Coordinates (*x*;*y*) (see :py:class:`Coordinates`)
    """

    def __init__(
        self,
        lunum: Optional["Lunum"] = None,
        coordinates: Optional["Coordinates"] = None,
    ) -> None:
        self._lunum = lunum
        self._coordinates = coordinates

    @property
    def lunum(self) -> "Lunum":
        """Return the LUNUM."""
        return self._lunum

    @property
    def coordinates(self) -> Union["Coordinates", None]:
        """Return the coordinates."""
        return self._coordinates

    def __str__(self):
        wire_anchor = []
        if self._coordinates:
            wire_anchor.append(str(self._coordinates))
        if self._lunum:
            wire_anchor.append(str(self._lunum))
        if len(wire_anchor) == 0:
            return ""
        return "|".join(wire_anchor)


class WirePath:
    """The wire path structure.

    Wire path is defined as:

    .. code-block:: ebnf

        path = {move} w_anchor
                | {move} branch

    *{move}* means zero or more moves.
    where:

    - :code:`move` is defined in :py:class:`Move`
    - :code:`w_anchor` is defined in :py:class:`WireAnchor`
    - :code:`branch` is defined in :py:class:`Branch`
    """

    def __init__(
        self,
        moves: Optional[List["Move"]] = None,
        w_anchor: Optional["WireAnchor"] = None,
        branch: Optional["Branch"] = None,
    ) -> None:
        self._moves = moves
        self._w_anchor = w_anchor
        self._branch = branch

    @property
    def moves(self) -> Union[List["Move"], None]:
        """Return the moves."""
        return self._moves

    @property
    def w_anchor(self) -> Union["WireAnchor", None]:
        """Return the wire anchor."""
        return self._w_anchor

    @property
    def branch(self) -> Union["Branch", None]:
        """Return the branch."""
        return self._branch

    def __str__(self) -> str:
        branch_str = []
        if self._moves:
            for move in self._moves:
                branch_str.append(str(move))
        if self._w_anchor:
            branch_str.append(str(self._w_anchor))
        if self._branch:
            branch_str.append(str(self._branch))
        if len(branch_str) == 0:
            return ""
        return " ".join(branch_str)


class Move:
    """Movement of a wire or transition.

    Move is defined as:

    .. Code-block:: ebnf

        Move = move_coordinates

    Where:

    - :code:`move_coordinates`: Move coordinates (see :py:class:`MoveCoordinates`)
    """

    def __init__(self, coordinates: Optional["MoveCoordinates"] = None) -> None:
        self._coordinates = coordinates

    @property
    def coordinates(self) -> Union["MoveCoordinates", None]:
        """Return the coordinates."""
        return self._coordinates

    def __str__(self) -> str:
        if not self._coordinates:
            return ""
        return str(self._coordinates)


class MoveCoordinates:
    """*Move coordinates* manages the movement among coordinates of wires and transitions.

    *Wire coordinates* are defined as:

    .. code-block:: ebnf

        wire_coordinates = COORD | coordinates

    and *transition coordinates* are defined as:

    .. code-block:: ebnf

        transition_coordinates = (COORD | coordinates) '|' coordinates

    Thus, joining both definitions, *move coordinates* is defined as:

    .. code-block:: ebnf

        move_coordinates = (COORD | coordinates) '|' coordinates
    """

    def __init__(
        self,
        coordinates: "Coordinates",
        fork_coordinates: Optional["Coordinates"] = None,
    ) -> None:
        self._coordinates = coordinates
        self._fork_coordinates = fork_coordinates

    @property
    def coordinates(self) -> "Coordinates":
        """Return the coordinates."""
        return self._coordinates

    @property
    def fork_coordinates(self) -> Union["Coordinates", None]:
        """Return the fork coordinates. Only applies to transitions."""
        return self._fork_coordinates

    def __str__(self):
        if not self._fork_coordinates:
            return str(self._coordinates)
        return f"{self._coordinates}|{self._fork_coordinates}"


class Branch:
    """Branch is the graphic path followed by the wire or transition.

    Branch is defined as:

    .. code-block:: ebnf

        branch = '[' path_list ']';
        path_list = path
            | path_list ',' path

    where:

    - :code:`path`: Wire path (see :py:class:`WirePath`)
    """

    def __init__(self, path_list: List["WirePath"]) -> None:
        self._path_list = path_list

    @property
    def path_list(self) -> List["WirePath"]:
        """Return the path list."""
        return self._path_list

    def __str__(self):
        path_list_str = []
        for path in self._path_list:
            path_list_str.append(str(path))
        if len(path_list_str) == 0:
            return ""
        list_str = ", ".join(path_list_str)
        return f"[{list_str}]"
